#!/usr/bin/env python3
"""
Test class for the Avro classes generator.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

import unittest
from pathlib import Path

from avro_preprocessor.avro_paths import AvroPaths

from pyavro_gen.generator import AvroGenerator
from pyavro_gen.pyavrogen_test import do_test_generated_classes

ROOT_DIR = Path(__file__).absolute().parent.parent


class AvroPreprocessorTest(unittest.TestCase):
    """
    Test class for the Avro schema extension.
    """

    @staticmethod
    def test_fixtures() -> None:
        """
        Standard test.
        """
        generator = AvroGenerator(
            AvroPaths(
                input_path=str(ROOT_DIR.joinpath('fixtures/')),
                output_path=str(ROOT_DIR.joinpath('avroclasses/')),
                input_schema_file_extension='avsc',
                base_namespace='com.jaumo.schema',
                rpc_namespace='com.jaumo.schema.rpc',
                types_namespace='com.jaumo.schema.type',
            )
        )

        generator.process()
        do_test_generated_classes(
            module=str(ROOT_DIR.joinpath('avroclasses/')),
            domain_namespace='com.jaumo.schema.domain',
            number_of_cycles=100
        )

    # @unittest.skip("fixture not available")
    @staticmethod
    def test_large_set() -> None:
        """
        Test on a large set of schemas.
        """

        generator = AvroGenerator(
            AvroPaths(
                input_path=str(ROOT_DIR.joinpath('../event_schema/build/schema/')),
                output_path=str(ROOT_DIR.joinpath('avroclasses/')),
                input_schema_file_extension='avsc',
                base_namespace='com.jaumo.message_schema',
                rpc_namespace='com.jaumo.message_schema.rpc',
                types_namespace='com.jaumo.message_schema.type',
            )
        )

        generator.process()
        do_test_generated_classes(str(ROOT_DIR.joinpath('avroclasses/')),
                                  'com.jaumo.message_schema.domain')



"""
Codewriter for Python modules
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from dataclasses import dataclass
from typing import List, Dict, Set

from isort import SortImports

from pyavro_gen.codewriters.core import ClassWriter, Dependency, WritingMode
from pyavro_gen.codewriters.utils import j, ind, to_snake_case


@dataclass
class ClassItem:
    """
    A dataclass to associate a class with its namespace, testing class, and testing class namespace.
    """
    namespace: str
    name: str
    namespace_test: str
    factory_name: str


class NamespaceWriter:
    """
    Codewriter for Python modules.
    It deals with stringifying all the classes.
    """

    def __init__(self, namespace: str):
        self.namespace = namespace
        self.class_writers: List[ClassWriter] = []

    def get_all_testing_classes(self) -> List[ClassItem]:
        """
        Returns a list of ClassItems to be written in the root module of the testing package
        to be generated.
        :return: A list of ClassItems
        """
        deps: List[Dependency] = [
            dep for writer in self.class_writers
            for dep in writer.as_dependency(WritingMode.GENERATION_INDEPENDENT)
        ]
        test_deps: List[Dependency] = [
            dep for writer in self.class_writers
            for dep in writer.as_dependency(WritingMode.TEST)
        ]

        return [
            ClassItem(
                j(dep.prefix, '.', dep.where),
                dep.what,
                j(test_deps[idx].prefix, '.', dep.where),
                test_deps[idx].what
            )
            for idx, dep in enumerate(deps)
        ]

    def write(self, mode: WritingMode = WritingMode.GENERATION) -> Dict[str, str]:
        """
        Returns this namespace as a Dict of str, using the specified writing mode.
        Keys are filenames; Values are file contents.

        :param mode: The writing mode
        :return: A dict of filenames-file contents
        """

        self.class_writers: List[ClassWriter] = \
            sorted(self.class_writers, key=lambda w: len(
                w.external_dependencies
            ))

        if mode == WritingMode.GENERATION:
            classes = (writer.write() for writer in self.class_writers)
            all_deps: Set[str] = set(
                dep.write()
                for writer in self.class_writers for dep in writer.all_dependencies(mode)
                if dep.where != self.namespace
            )

            content = j(
                j(*all_deps),
                j(*classes)
            )

            return {'__init__.py': SortImports(file_contents=content).output}

        # common content
        deps: List[str] = [
            dep.write(WritingMode.INIT)
            for writer in self.class_writers for dep in writer.as_dependency(mode)
        ]

        content = j(*deps)

        # NOTE: DO NOT SORT!
        #
        # It is important to let the interpreter read the files in the correct order.
        # E.g. the following won't work:
        # contents = {'__init__.py': SortImports(file_contents=content).output}
        contents = {'__init__.py': content}

        for writer in self.class_writers:
            if mode == WritingMode.GENERATION_INDEPENDENT:
                contents[f'{to_snake_case(writer.name)}.py'] = writer.write(mode)
            elif mode == WritingMode.TEST:
                contents[f'{to_snake_case(writer.name)}_factory.py'] = writer.write(mode)

        return contents


def get_testing_base() -> str:
    """
    Header for testing package __init__.py
    """
    return j(
        'import faker\n\n',
        'from pyavro_gen.enum_with_schema_provider import EnumWithSchemaProvider\n',
        'from .testing_classes import test_classes\n\n'
        'fake = faker.Faker()\n',
        'fake.add_provider(EnumWithSchemaProvider)\n',
    )


def get_testing_class_list(classes: List[ClassItem]) -> str:
    """
    Converts the list of testing classitems to a string.
    :param classes: The testing class items
    """
    return j(
        'from pyavro_gen.codewriters.namespace import ClassItem\n\n'
        'test_classes = [\n',
        *(
            f"{ ind(1)}ClassItem("
            f"'{c.namespace}', '{c.name}', '{c.namespace_test}', '{c.factory_name}'"
            f"),\n"
            for c in classes
        ),
        ']\n'
    )



"""
Extended codewriters.
"""

from typing import Optional

from pyavro_gen.codewriters.core import ClassWriter, Decorator, Extension, Method, Docstring, \
    Attribute, SimpleTypeClassWriter
from pyavro_gen.codewriters.utils import j

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class UndictifiableClassWriter(ClassWriter):
    """
    Codewriter to produce classes that can be "undictified".
    https://github.com/Dobiasd/undictify
    """

    def __init__(self,
                 fully_qualified_name: str,
                 doc: Optional[str] = None,
                 prefix: Optional[str] = None):
        super().__init__(fully_qualified_name, doc, prefix)

        self.decorators = [
            Decorator('@type_checked_constructor()',
                      ClassWriter('undictify.type_checked_constructor')),
            Decorator('@dataclass', ClassWriter('dataclasses.dataclass'))
        ]

        self.methods.append(Method(
            name='to_dict',
            parameters=[Attribute('self')],
            return_type=ClassWriter('typing.Dict'),
            doc=Docstring('Returns a dictionary version of this instance.'),
            body="return asdict(self)",
            additional_dependencies=[ClassWriter('dataclasses.asdict')]
        ))

        self.methods.append(Method(
            name='from_dict',
            parameters=[
                Attribute('cls'),
                Attribute(
                    name='the_dict',
                    type=ClassWriter('typing.Dict'),
                    doc=Docstring('The dictionary from which to create an instance of this class.')
                )
            ],
            return_type=SimpleTypeClassWriter(j("'", self.name, "'")),  # simple: no dep added
            doc=Docstring('Returns an instance of this class from a dictionary.'),
            body="return cls(**the_dict)",
            decorators=[Decorator('@classmethod')],
        ))


class DataClassWriter(ClassWriter):
    """
    Codewriter to write dataclasses.
    """

    def __init__(self,
                 fully_qualified_name: str,
                 doc: Optional[str] = None,
                 prefix: Optional[str] = None):
        super().__init__(fully_qualified_name, doc, prefix)

        self.decorators = [
            Decorator('@dataclass', ClassWriter('dataclasses.dataclass'))
        ]


class RpcWriter(ClassWriter):
    """
    Codewriter to write RPC protocols.
    """

    def __init__(self,
                 fully_qualified_name: str,
                 doc: Optional[str] = None,
                 prefix: Optional[str] = None):
        super().__init__(fully_qualified_name, doc, prefix)

        self.extensions = [
            Extension('abc.ABC')
        ]



"""
A provider for enums with schemas
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from enum import Enum
from typing import TypeVar, Type, List, Iterable, cast

from faker.providers import BaseProvider

TEnum = TypeVar("TEnum", bound=Enum)


class EnumWithSchemaProvider(BaseProvider):  # type: ignore
    """
    A Provider for enums with schemas.
    """

    def enum_with_schema(self, enum_cls: Type[TEnum]) -> TEnum:
        """
        Return a random element of this enum

        :param enum_cls: Enum class
        """
        members: List[TEnum] = list(m for m in cast(Iterable[TEnum], enum_cls)
                                    if not m.name.startswith('_'))
        return cast(TEnum, self.random_element(members))



#!/usr/bin/env python
#
# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


#
# derived from https://github.com/verisign/python-confluent-schemaregistry.git
#

"""
Copy/paste from confluent-kafka-python tests
https://github.com/confluentinc/confluent-kafka-python/blob/master/tests/avro/mock_schema_registry_client.py

latest update: 12/02/2019
"""


from confluent_kafka.avro import ClientError


class MockSchemaRegistryClient(object):
    """
    A client that acts as a schema registry locally.

    Compatibiity related methods are not implemented at this time.
    """

    def __init__(self, max_schemas_per_subject=1000):
        self.max_schemas_per_subject = max_schemas_per_subject
        # subj => { schema => id }
        self.subject_to_schema_ids = {}
        # id => avro_schema
        self.id_to_schema = {}
        # subj => { schema => version }
        self.subject_to_schema_versions = {}

        self.subject_to_latest_schema = {}

        # counters
        self.next_id = 1
        self.schema_to_id = {}

    def _get_next_id(self, schema):
        if schema in self.schema_to_id:
            return self.schema_to_id[schema]
        result = self.next_id
        self.next_id += 1
        self.schema_to_id[schema] = result
        return result

    def _get_next_version(self, subject):
        if subject not in self.subject_to_schema_versions:
            self.subject_to_schema_versions[subject] = {}
        return len(self.subject_to_schema_versions[subject])

    def _get_all_versions(self, subject):
        versions = self.subject_to_schema_versions.get(subject, {})
        return sorted(versions)

    def _add_to_cache(self, cache, subject, schema, value):
        if subject not in cache:
            cache[subject] = {}
        sub_cache = cache[subject]
        sub_cache[schema] = value

    def _cache_schema(self, schema, schema_id, subject, version):
        # don't overwrite anything
        if schema_id in self.id_to_schema:
            schema = self.id_to_schema[schema_id]
        else:
            self.id_to_schema[schema_id] = schema

        self._add_to_cache(self.subject_to_schema_ids,
                           subject, schema, schema_id)

        self._add_to_cache(self.subject_to_schema_versions,
                           subject, schema, version)

        if subject in self.subject_to_latest_schema:
            si, s, v = self.subject_to_latest_schema[subject]
            if v > version:
                return
        self.subject_to_latest_schema[subject] = (schema_id, schema, version)

    def register(self, subject, avro_schema):
        """
        Register a schema with the registry under the given subject
        and receive a schema id.

        avro_schema must be a parsed schema from the python avro library

        Multiple instances of the same schema will result in inconsistencies.
        """
        schemas_to_id = self.subject_to_schema_ids.get(subject, {})
        schema_id = schemas_to_id.get(avro_schema, -1)
        if schema_id != -1:
            return schema_id

        # add it
        version = self._get_next_version(subject)
        schema_id = self._get_next_id(avro_schema)

        # cache it
        self._cache_schema(avro_schema, schema_id, subject, version)
        return schema_id

    def get_by_id(self, schema_id):
        """Retrieve a parsed avro schema by id or None if not found"""
        return self.id_to_schema.get(schema_id, None)

    def get_latest_schema(self, subject):
        """
        Return the latest 3-tuple of:
        (the schema id, the parsed avro schema, the schema version)
        for a particular subject.

        If the subject is not found, (None,None,None) is returned.
        """
        return self.subject_to_latest_schema.get(subject, (None, None, None))

    def get_version(self, subject, avro_schema):
        """
        Get the version of a schema for a given subject.

        Returns -1 if not found.
        """
        schemas_to_version = self.subject_to_schema_versions.get(subject, {})
        return schemas_to_version.get(avro_schema, -1)

    def get_id_for_schema(self, subject, avro_schema):
        """
        Get the ID of a parsed schema
        """
        schemas_to_id = self.subject_to_schema_ids.get(subject, {})
        return schemas_to_id.get(avro_schema, -1)

    def test_compatibility(self, subject, avro_schema, version='latest'):
        raise ClientError("not implemented")

    def update_compatibility(self, level, subject=None):
        raise ClientError("not implemented")

    def get_compatibility(self, subject=None):
        raise ClientError("not implemented")



#!/usr/bin/env python

"""
This scripts tests generated classes using Kafka Avro Serde.
"""
import argparse
import importlib
import sys
from dataclasses import asdict
from enum import Enum
from pathlib import Path

from avro.schema import Parse, Schema
from confluent_kafka.avro.serializer import TopicRecordNameStrategy

from pyavro_gen.codewriters.namespace import ClassItem
from pyavro_gen.codewriters.utils import j
from pyavro_gen.equality_test import are_almost_equal
from pyavro_gen.mock_schema_registry_client import MockSchemaRegistryClient

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


def do_test_generated_classes(  # pylint: disable=R0914
        module: str,
        domain_namespace: str,
        number_of_cycles: int = 10,
        verbose: bool = True
) -> None:
    """
    Test generated classes. For each generated class that corresponds to an Avro record,
    this function performs a serialization/deserialization cycle and checks if input and output
    are equal.

    :param module: The path of the package of the generated classes
    :param domain_namespace: The namespace of Avro "root" classes that can be a ser/deserialised.
    :param number_of_cycles: How many serde cycles to perform per class
    :param verbose: To show any output or not
    """

    # To be able to load a package from an arbitrary location, we add the path sys.path
    # we perform cleanup later
    module_path = Path(module).absolute()
    parent_path = module_path.parent
    sys.path.insert(0, str(parent_path))
    module_name = str(module_path.name)
    test_module_name = module_name + '_test'

    classes_module = importlib.import_module(module_name)
    test_module = importlib.import_module(test_module_name)

    # reload useful for bulk unittests
    importlib.reload(classes_module)
    test_module = importlib.reload(test_module)
    test_module.testing_classes = importlib.reload(test_module.testing_classes)  # type: ignore

    schema_registry_client = MockSchemaRegistryClient()  # type: ignore

    value_serializer = classes_module.TypedAvroSerializer(  # type: ignore
        schema_registry_client,
        is_key=False,
        subject_strategy=TopicRecordNameStrategy
    )

    value_deserializer = classes_module.TypedAvroDeserializer(  # type: ignore
        schema_registry_client,
        is_key=False
    )

    for test_class_item in test_module.testing_classes.test_classes:  # type: ignore
        class_item: ClassItem = test_class_item

        if verbose:
            print('testing', j(class_item.namespace, '.', class_item.name))

        mod = __import__(class_item.namespace, fromlist=[class_item.name])
        klass = getattr(mod, class_item.name)

        test_mod = __import__(class_item.namespace_test, fromlist=[class_item.factory_name])
        klass_factory = getattr(test_mod, class_item.factory_name)

        if domain_namespace not in class_item.namespace or issubclass(klass, Enum):
            # Other classes cannot be directly serialised by Avro so we skip them
            if verbose:
                print('\tskipping')
            continue

        # we only check records with a full namespace
        avro_schema: Schema = Parse(klass._schema)  # pylint: disable=W0212
        if avro_schema.fullname.startswith('.'):
            continue

        for i in range(number_of_cycles):
            instance = klass_factory()
            if verbose:
                print(i, '\t', instance)

            serialized = value_serializer('mytopic', instance)
            deserialized = value_deserializer('mytopic', serialized)

            try:
                assert are_almost_equal(
                    instance, deserialized, max_abs_ratio_diff=1.000001, max_abs_diff=-1)
            except AssertionError:
                print("expected:", instance)
                print("actual:  ", deserialized)
                print("des dict:", asdict(deserialized))
                clean_path()
                raise

    clean_path()


def clean_path() -> None:
    """
    Removes the path added in do_test_generated_classes()
    """
    sys.path.pop(0)


if __name__ == '__main__':
    PARSER = argparse.ArgumentParser(
        description='A test for Avro typed classes generated by pyavro-gen.')

    PARSER.add_argument('-m', '--module', dest='module', type=str, required=True)
    PARSER.add_argument('-n', '--number_of_cycles', dest='ncycles', type=int, default=10)
    PARSER.add_argument('-d', '--domain-namespace', dest='domain_namespace', type=str,
                        required=True)
    PARSER.add_argument('-v', '--verbose', dest='verbose', action='store_true')
    PARSER.set_defaults(verbose=True)

    ARGS = PARSER.parse_args()
    do_test_generated_classes(ARGS.module, ARGS.domain_namespace, ARGS.ncycles, ARGS.verbose)



"""
This module defines the class that specifies the types of generated classes
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from enum import Enum
from typing import Type, Dict, Any

from pyavro_gen.codewriters.core import ClassVarClassWriter, EnumClassWriter
from pyavro_gen.codewriters.extended import UndictifiableClassWriter, RpcWriter


class GenerationClassesType(Enum):
    """
    An enum for types of generation classes
    """
    RECORD_CLASS = 'RECORD_CLASS'
    RPC_CLASS = 'RPC_CLASS'
    CLASS_VARIABLE_CLASS = 'CLASS_VARIABLE_CLASS'
    ENUM_CLASS = 'ENUM_CLASS'


#: Container for types of generation classes. Values should be codewriters.
GENERATION_CLASSES: Dict[GenerationClassesType, Type[Any]] = {
    GenerationClassesType.RECORD_CLASS: UndictifiableClassWriter,
    GenerationClassesType.RPC_CLASS: RpcWriter,
    GenerationClassesType.CLASS_VARIABLE_CLASS: ClassVarClassWriter,
    GenerationClassesType.ENUM_CLASS: EnumClassWriter,
}



"""
A module to collect fields from Avro schemas to generate Python classes.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

import json
from collections import OrderedDict
from typing import List

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule

from pyavro_gen.codewriters.containers import UnionClassWriter, ListClassWriter, MapClassWriter
from pyavro_gen.codewriters.core import ClassWriter, Attribute, Method, EnumValueWriter, Comment, \
    Docstring, SimpleTypeClassWriter
from pyavro_gen.codewriters.logicaltypes import DatetimeClassWriter, DateClassWriter, \
    UUIDClassWriter
from pyavro_gen.codewriters.utils import namespace_name
from pyavro_gen.generation_classes import GENERATION_CLASSES, GenerationClassesType
from pyavro_gen.modules.avsc_schema_dependency_checker import AvscSchemaDependenciesChecker
from pyavro_gen.schema_and_classes_container import SchemaAndClassesContainer


class FieldsCollector(PreprocessorModule):  # type: ignore
    """
    Collects fields to construct classes in all schemas.
    """

    def __init__(self, schemas: SchemaAndClassesContainer) -> None:
        super().__init__(schemas)

        self.writers = schemas.output_writers

        self.prefix = schemas.output_prefix

    def process(self) -> None:
        """Process all schemas."""

        for _, schema in self.processed_schemas_iter():
            self.traverse_schema(self.fill_fields_of_node, schema)

    def fill_fields_of_node(self, node: Avro.Node) -> None:
        """
        Find fields in an Avro Node.
        :param node: The node
        """
        if not isinstance(node, OrderedDict):
            return

        if Avro.Protocol in node:
            self.fill_fields_of_protocol_record(node)

        elif Avro.Name in node and (Avro.Fields in node or Avro.Symbols in node):
            self.fill_fields_of_standard_record(node)

    # pylint: disable=R0914,R0912
    def fill_fields_of_protocol_record(self, node: Avro.Node) -> None:
        """
        Fills attributes, methods, and documentation of a ClassWriter for Avro protocol records.
        :param node: The Avro node
        """
        writer = self.get_class_writer_from_name(
            self.current_schema_name, GenerationClassesType.RPC_CLASS
        )

        if Avro.Doc in node:
            writer.doc = Docstring(node[Avro.Doc])

        for message_name, message in node[Avro.Messages].items():
            request_fields: List[Attribute] = []
            for field in message[Avro.Request]:
                avro_type = field[Avro.Type]
                field_type = self.get_class_writer_for_node(avro_type)
                field_type.prefix = self.prefix
                request_field = Attribute(field[Avro.Name], field_type)
                if Avro.Doc in field:
                    request_field.doc = Docstring(field[Avro.Doc])
                request_fields.append(request_field)

            doc = message.get(Avro.Doc, None)

            response = message[Avro.Response]
            if isinstance(response, OrderedDict):
                if Avro.Namespace in response:
                    response_fully_qualified_name = \
                        response[Avro.Namespace] + '.' + response[Avro.Name]
                else:
                    response_fully_qualified_name = \
                        namespace_name(self.current_schema_name) + '.' + response[Avro.Name]

            else:
                # response is string
                if response == 'null' or Avro.OneWay:
                    response_fully_qualified_name = 'None'
                elif '.' in response[Avro.Name]:
                    response_fully_qualified_name = response[Avro.Name]
                else:
                    raise TypeError(f'Type unsupported: {response}')

            errors = message.get(Avro.Errors)
            if errors:
                errors_writers = []
                for error in errors:
                    if '.' in error:
                        err = self.get_class_writer_from_name(error)
                    else:
                        err = self.get_class_writer_from_name(
                            f'{namespace_name(self.current_schema_name)}.{error}')
                    errors_writers.append(err)
            else:
                errors_writers = []

            writer.methods.append(Method(
                name=message_name,
                parameters=[Attribute('self')] + request_fields,
                return_type=self.get_class_writer_from_name(response_fully_qualified_name),
                doc=Docstring(doc),
                errors=errors_writers
            ))

    def fill_fields_of_standard_record(self, node: Avro.Node) -> None:
        """
        Fills attributes and documentation of a ClassWriter for Avro records.
        :param node: The Avro node
        """
        if Avro.Namespace in node:
            fully_qualified_name = node[Avro.Namespace] + '.' + node[Avro.Name]
        elif '.' in node[Avro.Name]:
            fully_qualified_name = node[Avro.Name]
        else:
            fully_qualified_name = \
                namespace_name(self.current_schema_name) + '.' + node[Avro.Name]

        writer = self.get_class_writer_from_name(fully_qualified_name)

        if not writer.attributes:
            if Avro.Fields in node:
                for field in node[Avro.Fields]:
                    avro_type = field[Avro.Type]
                    field_type = self.get_class_writer_for_node(avro_type)
                    field_type.prefix = self.prefix
                    attribute = Attribute(field[Avro.Name], field_type)
                    if Avro.Doc in field:
                        attribute.doc = Docstring(field[Avro.Doc])
                    writer.attributes.append(attribute)
            else:
                fully_qualified_name = writer.namespace + '.' + writer.name
                writer = self.get_class_writer_from_name(
                    fully_qualified_name, GenerationClassesType.ENUM_CLASS)

                for field in node[Avro.Symbols]:
                    writer.attributes.append(Attribute(field, EnumValueWriter(field)))

            writer.attributes.append(
                Attribute(
                    name='_schema',
                    type=GENERATION_CLASSES[GenerationClassesType.CLASS_VARIABLE_CLASS](
                        SimpleTypeClassWriter('str')),
                    doc=Docstring('The Avro Schema associated to this class'),
                    default_value=f'"""{json.dumps(node, indent=4)}"""'
                )
            )

        if Avro.Doc in node:
            writer.doc = Docstring(node[Avro.Doc])

    def get_class_writer_from_name(
            self,
            fully_qualified_name: str,
            writer_type: GenerationClassesType = GenerationClassesType.RECORD_CLASS
    ) -> ClassWriter:
        """
        Returns a Class writer for the record in input. If it not exists, this method creates one.

        :param fully_qualified_name: The fully qualified name of the record
        :param writer_type: The type of writer
        """
        schemas: SchemaAndClassesContainer = self.schemas
        record_dependencies: 'OrderedDict[str, List[str]]' = \
            schemas.modules[AvscSchemaDependenciesChecker.__name__].record_dependencies

        if fully_qualified_name in record_dependencies:
            if fully_qualified_name in self.writers:
                if isinstance(self.writers[fully_qualified_name], GENERATION_CLASSES[writer_type]):
                    return self.writers[fully_qualified_name]
                # if it is not, we create and store a new one of the requested type

            writer: ClassWriter = GENERATION_CLASSES[writer_type](
                fully_qualified_name,
                prefix=schemas.output_prefix
            )

            writer.external_dependencies = record_dependencies[fully_qualified_name]
            self.writers[fully_qualified_name] = writer
            return writer

        return SimpleTypeClassWriter(fully_qualified_name)

    # pylint: disable=R0915,R0912
    def get_class_writer_for_node(self, avro_type: Avro.Node) -> ClassWriter:
        """
        Recursively returns an appropriate class writer given an Avro node
        :param avro_type: An Avro Node
        """

        # to make mypy happy
        field_type: ClassWriter

        if avro_type == 'boolean':
            field_type = SimpleTypeClassWriter('bool')
        elif avro_type == 'string':
            field_type = SimpleTypeClassWriter('str')
        elif avro_type == 'int':
            field_type = SimpleTypeClassWriter('int')
        elif avro_type == 'long':
            field_type = SimpleTypeClassWriter('int')
        elif avro_type == 'float':
            field_type = SimpleTypeClassWriter('float')
        elif avro_type == 'double':
            field_type = SimpleTypeClassWriter('float')
        elif avro_type == 'null':
            field_type = SimpleTypeClassWriter('None')
        elif isinstance(avro_type, str):
            if '.' in avro_type:
                field_type = self.get_class_writer_from_name(avro_type)
            else:
                raise ValueError(f'Type unsupported: {str(avro_type)}')

        elif isinstance(avro_type, list):
            field_type = UnionClassWriter([self.get_class_writer_for_node(t) for t in avro_type])
        elif isinstance(avro_type, OrderedDict):
            if Avro.Type in avro_type:
                if avro_type[Avro.Type] in [Avro.Record, Avro.Enum]:
                    if Avro.Namespace in avro_type:
                        name = avro_type[Avro.Namespace] + '.' + avro_type[Avro.Name]
                    else:
                        if '.' in avro_type[Avro.Name]:
                            name = avro_type[Avro.Name]
                        else:
                            name = namespace_name(self.current_schema_name) \
                                   + '.' + avro_type[Avro.Name]
                    field_type = self.get_class_writer_from_name(name)
                elif avro_type[Avro.Type] == Avro.Array:
                    field_type = ListClassWriter(
                        self.get_class_writer_for_node(avro_type[Avro.Items]))
                elif avro_type[Avro.Type] == Avro.Map:
                    field_type = MapClassWriter(
                        self.get_class_writer_for_node(avro_type[Avro.Values]))
                elif Avro.LogicalType in avro_type:
                    if avro_type[Avro.LogicalType] == 'timestamp-micros':
                        field_type = DatetimeClassWriter()
                    elif avro_type[Avro.LogicalType] == 'timestamp-millis':
                        field_type = DatetimeClassWriter('milliseconds')
                    elif avro_type[Avro.LogicalType] == 'date':
                        field_type = DateClassWriter()
                    elif avro_type[Avro.LogicalType] == 'uuid':
                        field_type = UUIDClassWriter()
                    else:
                        field_type = self.get_class_writer_for_node(avro_type[Avro.Type])
                    field_type.comment = Comment(f"logicalType: {avro_type[Avro.LogicalType]}")

                else:
                    raise TypeError(f'Type unsupported: {str(avro_type)}')
            else:
                raise TypeError(f'Type unsupported: {str(avro_type)}')
        else:
            raise TypeError(f'Type unsupported: {str(avro_type)}')

        return field_type



"""
The setup configuration.
"""

from pathlib import Path

from setuptools import setup, find_packages

setup(
    name='pyavro-gen',
    version_format='{tag}',
    description='A typed class generator for Avro Schemata',
    long_description=Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    keywords=['avro', 'classes', 'typing', 'types', 'type', 'typed', 'generation', 'creation',
              'schema', 'schemas', 'schemata'],
    url='https://gitlab.com/Jaumo/pyavro-gen',
    author='Jaumo GmbH',
    author_email='nicola.bova@jaumo.com',
    packages=find_packages(),
    scripts=[
        'pyavro_gen/pyavrogen.py',
        'pyavro_gen/pyavrogen_test.py'
    ],
    license='Apache2',
    install_requires=[
        'networkx==2.2',
        'pygments==2.2.0',
        'avro-python3',
        'pathlib',
        'factory_boy',
        'undictify>=0.6.2',
        'faker',
        'fastavro>=0.21.18',
        'isort',
        'avro-python3',
        'avro-preprocessor>=0.0.50',
    ],
)



"""
This module is added as __init__.py of the main packages of generated classes.

It patches fastavro enum reader and writer to take enums as input (and return as output)
instead of strings.

Since it cannot know which enum class to return, we keep a dictionary of enum classes
found in the package and match them with the fully qualified domain of the avro schema,
during deserialization.
"""
import glob
import inspect
import struct
from importlib import import_module
from pathlib import Path
from typing import Dict, Optional, Union, Any, cast, Callable

import confluent_kafka.avro.serializer
import fastavro._read_py as fastavro_read  # pylint: disable=W0212
import fastavro._validation_py as fastavro_validation  # pylint: disable=W0212
import fastavro._write_py as fastavro_write  # pylint: disable=W0212
from avro.schema import Schema
from confluent_kafka.avro import CachedSchemaRegistryClient, loads  # pylint: disable=C0412
# pylint: disable=C0412
from confluent_kafka.avro.serializer import AvroSerializer, TopicNameStrategy, \
    AvroDeserializer, SerializerError, ContextStringIO, MAGIC_BYTE  # pylint: disable=W0212,C0412
from fastavro._read_common import SchemaResolutionError  # pylint: disable=W0212
from fastavro._read_py import read_long  # pylint: disable=W0212
from fastavro._write_py import write_int  # pylint: disable=W0212

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

BASE_PACKAGE = '%%%base_package%%%'  # will be replaced

CLASSES: Dict[str, type] = {}

RESOLVED_NAMESPACE = Path(__file__).parent.resolve()
PACKAGE = RESOLVED_NAMESPACE.name.replace('/', '')
ROOT_DIR = str(RESOLVED_NAMESPACE.parent)
if not ROOT_DIR.endswith('/'):
    ROOT_DIR += '/'

for file in glob.iglob(str(RESOLVED_NAMESPACE) + '/**/__init__.py', recursive=True):
    module_name = str(Path(file).parent).replace(ROOT_DIR, '').replace('/', '.')
    module = import_module(module_name, PACKAGE)
    for _, class_type in inspect.getmembers(module, inspect.isclass):
        # Here we remove the base package.
        # Also, class module name is snake case, but classes are imported
        # in their packages' __init__.py so classes' module names have to be removed.
        # e.g.
        # from   'avroclasses.com.jaumo.schema.domain.user.gender.Gender'
        # to     'com.jaumo.schema.domain.user.Gender'
        class_module = class_type.__module__
        class_module_no_package = class_module[class_module.find('.') + 1:]
        if not class_module_no_package.startswith(BASE_PACKAGE):
            continue
        parent_class_module_no_package = \
            class_module_no_package[:class_module_no_package.rfind('.') + 1]
        fully_qualified_name = parent_class_module_no_package + class_type.__name__

        CLASSES[fully_qualified_name] = class_type
for k, v in CLASSES.items():
    print('loading class', k, v)


# pylint: disable=W0613
def my_validate_enum(datum, schema, **kwargs):  # type: ignore
    """
    Check that the data value matches one of the enum symbols.

    i.e "blue" in ["red", green", "blue"]

    Parameters
    ----------
    datum: Any
        Data being validated
    schema: dict
        Schema
    kwargs: Any
        Unused kwargs
    """
    return datum.value in schema['symbols']


# pylint: disable=C0103
def my_write_enum(fo, datum, schema):  # type: ignore
    """An enum is encoded by a int, representing the zero-based position of
    the symbol in the schema."""
    index = schema['symbols'].index(datum.value)
    write_int(fo, index)


# pylint: disable=C0103
def my_read_enum(fo, writer_schema, reader_schema=None):  # type: ignore
    """An enum is encoded by a int, representing the zero-based position of the
    symbol in the schema.
    """
    index = read_long(fo)
    symbol = writer_schema['symbols'][index]
    if reader_schema and symbol not in reader_schema['symbols']:
        default = reader_schema.get("default")
        if default:
            return default

        symlist = reader_schema['symbols']
        msg = '%s not found in reader symbol list %s' % (symbol, symlist)
        raise SchemaResolutionError(msg)

    klass_fully_qualified_name = writer_schema['name']
    klass = CLASSES[klass_fully_qualified_name]
    return klass(symbol)


fastavro_validation.VALIDATORS['enum'] = my_validate_enum
fastavro_write.WRITERS['enum'] = my_write_enum
fastavro_write.validate = fastavro_validation.validate
fastavro_read.READERS['enum'] = my_read_enum

confluent_kafka.avro.serializer.schemaless_writer = fastavro_write.schemaless_writer
confluent_kafka.avro.serializer.schemaless_reader = fastavro_read.schemaless_reader


# pylint: disable=R0903
class TypedAvroSerializer(AvroSerializer):  # type: ignore
    """
    Encodes kafka messages as Avro; registering the schema with the Confluent Schema Registry.

    :param registry_client CachedSchemaRegistryClient: Instance of CachedSchemaRegistryClient.
    :param bool is_key: True if configured as a key_serializer.
    :param func(str, bool, schema): Returns the subject name used when registering schemas.
    """

    def __init__(
            self,
            registry_client: str,
            is_key: bool = False,
            subject_strategy: Callable[[str, bool, Schema], str] = TopicNameStrategy
    ) -> None:
        super().__init__(registry_client, is_key, subject_strategy)

        self.class_name_to_schema: Dict[str, Schema] = {}

    def _get_parsed_schema(self, record: Any) -> Schema:
        klass = record.__class__
        class_name = klass.__module__ + '.' + klass.__name__
        if class_name in self.class_name_to_schema:
            return self.class_name_to_schema[class_name]

        parsed_schema = loads(record._schema)  # pylint: disable=W0212
        self.class_name_to_schema[class_name] = parsed_schema
        return parsed_schema

    def __call__(self, topic: str, record: Any) -> Optional[bytes]:
        """
        Given a parsed avro schema, encode a record for the given topic.

        The schema is registered with the subject of 'topic-value'.
        :param str topic: Topic name.
        :param Any record: An object to serialize.
        :returns: Encoded record with schema ID as bytes.
        :rtype: bytes
        """
        if record is None:
            return None

        schema = self._get_parsed_schema(record)

        subject = self.subject_strategy(topic, self.is_key, schema)

        schema_id = self.registry_client.register(subject, schema)
        if not schema_id:
            message = "Failed to register schema with subject {}.".format(subject)
            raise SerializerError(message, is_key=self.is_key)

        if schema_id not in self.codec_cache:
            self.codec_cache[schema_id] = self._get_encoder_func(schema)

        return cast(bytes, self._encode(schema_id, record.to_dict()))


# pylint: disable=R0903
class TypedAvroDeserializer(AvroDeserializer):  # type: ignore
    """
    Decodes Kafka messages encoded by Confluent Schema Registry compliant Avro Serializers.

    :param registry_client CachedSchemaRegistryClient: Instance of CachedSchemaRegistryClient.
    :param bool is_key: True if configured as a key_serializer.
    :param schema reader_schema: Optional reader schema to be used during deserialization.
    """

    def __init__(
            self,
            registry_client: CachedSchemaRegistryClient,
            is_key: bool = False,
            reader_schema: Optional[Schema] = None
    ) -> None:
        super().__init__(registry_client, is_key, reader_schema)

        self.schema_id_to_class: Dict[int, type] = {}

    def _get_class(self, schema_id: int) -> type:
        if schema_id in self.schema_id_to_class:
            return self.schema_id_to_class[schema_id]

        writer_schema: Schema = self.registry_client.get_by_id(schema_id)
        name = writer_schema.fullname
        self.schema_id_to_class[name] = CLASSES[name]
        return self.schema_id_to_class[name]

    def __call__(self, topic: str, datum: Union[str, bytes, None]) -> Any:
        """
        Decode a datum from kafka that has been encoded for use with the Confluent Schema Registry.
        :param str|bytes or None datum: message key or value to be decoded.
        :returns: Typed version of the decoded message key or value contents.
        """

        if datum is None:
            return None

        if len(datum) <= 5:
            raise SerializerError("message is too small to decode")

        with ContextStringIO(datum) as payload:
            magic, schema_id = struct.unpack('>bI', payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte", self.is_key)

            decoder_func = self._get_decoder_func(schema_id, payload)
            return self._get_class(schema_id).from_dict(decoder_func(payload))  # type: ignore
